# Copyright (c) Stephan Martin <sm@sm-zone.net>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.

use strict;

package OpenSSL;

sub new {
   my $self  = {};
   my $that  = shift;
   my $main  = shift;
   my $class = ref($that) || $that;

   $self->{'bin'} = $main->{'init'}->{'opensslbin'};
   $main->print_error("Can't execute OpenSSL: $self->{'bin'}") 
      if (! -x $self->{'bin'});

   $self->{'tmp'}  = $main->{'init'}->{'basedir'}."/tmp";

   bless($self, $class);
}

sub newkey {
   my $self = shift;
   my $opts = { @_ };

   my $cmd = "$self->{'bin'} genrsa";
   $cmd .= " -des3";
   $cmd .= " -passout env:SSLPASS";

   $cmd .= " -out $opts->{'of'}";

   $cmd .= " $opts->{'bits'}";
   $ENV{'SSLPASS'} = $opts->{'pw'};
   system($cmd);
   delete($ENV{'SSLPASS'});
}

sub signreq {
   my $self = shift;
   my $opts = { @_ };

   my $cmd = "$self->{'bin'} ca -batch";
   $cmd .= " -passin env:SSLPASS";

   $cmd .= " -config $opts->{'cf'}";
   $cmd .= " -name $opts->{'cn'}" if($opts->{'cn'} ne "");
   $cmd .= " -in $opts->{'rf'}";
   $cmd .= " -days $opts->{'ds'}";
   $cmd .= " -preserveDN";

   $ENV{'SSLPASS'} = $opts->{'pw'};
   #print "DEBUG call: $cmd\n";
   system("$cmd");
   delete($ENV{'SSLPASS'});

   return;
}

sub revoke {
   my $self = shift;
   my $opts = { @_ };

   my $cmd = "$self->{'bin'} ca";
   $cmd .= " -passin env:SSLPASS";

   $cmd .= " -config $opts->{'cf'}";
   $cmd .= " -revoke $opts->{'if'}";

   $ENV{'SSLPASS'} = $opts->{'pw'};
   # print "DEBUG call: $cmd\n";
   system("$cmd") && return(1);
   delete($ENV{'SSLPASS'});

   return;
}

sub newreq {
   my $self = shift;
   my $opts = { @_ };

   my $cmd = "$self->{'bin'} req -new";
   $cmd .= " -keyform PEM";
   $cmd .= " -outform PEM";
   $cmd .= " -passin env:SSLPASS";

   $cmd .= " -config $opts->{'cf'}";
   $cmd .= " -out $opts->{'of'}";
   $cmd .= " -key $opts->{'kf'}";

   $ENV{'SSLPASS'} = $opts->{'pw'};
   #   print "DEBUG call: $cmd\n";
   open(CMD, "|$cmd");
   print CMD "$_\n" foreach(@{$opts->{'dn'}});
   close CMD;
   delete($ENV{'SSLPASS'});

   return;
}

sub newcert {
   my $self = shift;
   my $opts = { @_ };

   my $cmd = "$self->{'bin'} req -x509";
   $cmd .= " -keyform PEM";
   $cmd .= " -outform PEM";
   $cmd .= " -passin env:SSLPASS";

   $cmd .= " -config $opts->{'cf'}";
   $cmd .= " -out $opts->{'of'}";
   $cmd .= " -key $opts->{'kf'}";
   $cmd .= " -in $opts->{'rf'}";
   $cmd .= " -days $opts->{'ds'}";

   $ENV{'SSLPASS'} = $opts->{'pw'};
   system($cmd);
   delete($ENV{'SSLPASS'});
}

sub newcrl {
   my $self = shift;
   my $opts = { @_ };

   my $out;

   my $tmpfile = _mktmp($self->{'tmp'}."/crl");
   my $cmd = "$self->{'bin'} ca -gencrl";
   $cmd .= " -passin env:SSLPASS";
   $cmd .= " -config $opts->{'conf'}";

   $cmd .= " -out $tmpfile";
   $cmd .= " -crldays $opts->{'days'}";

   $ENV{'SSLPASS'} = $opts->{'pass'};
   system($cmd);
   delete($ENV{'SSLPASS'});

   my $crl = $self->parsecrl('', $tmpfile, 1);

   if($opts->{'format'} eq 'PEM') {
      $out = $crl->{'PEM'};
   } elsif ($opts->{'format'} eq 'DER') {
      $out = $crl->{'DER'};
   } elsif ($opts->{'format'} eq 'TEXT') {
      $out = $crl->{'TEXT'};
   } else {
      $out = $crl->{'PEM'};
   }

   open(OUT, ">$opts->{'outfile'}");
   print OUT $out;
   close OUT;
   
}
   
sub parsecrl {
   my $self  = shift;
   my $main  = shift;
   my $file  = shift;
   my $force = shift;

   my $tmp   = {};
   my (@lines, $i);

   # check if crl is cached
   if($self->{'CACHE'}->{$file} && not $force) {
      return($self->{'CACHE'}->{$file});
   }

   open(IN, $file) || do {
      $main->print_warning("Can't open crl: $!");
      return;
   };

   # convert crl to PEM, DER and TEXT
   $tmp->{'PEM'} .= $_ while(<IN>);
   $tmp->{'TEXT'} = $self->_convcrl(
         $main,
         $tmp->{'PEM'},
         'TEXT'
         );
   $tmp->{'DER'} = $self->_convcrl(
         $main,
         $tmp->{'PEM'},
         'DER'
         );

   # get "normal infos"
   @lines = split(/\n/, $tmp->{'TEXT'});
   foreach(@lines) {
      if ($_ =~ /Signature Algorithm.*: (\w+)/i) {
         $tmp->{'SIG_ALGORITHM'} = $1;
      } elsif ($_ =~ /Issuer: (.+)/i) {
         $tmp->{'ISSUER'} = $1;
         $tmp->{'ISSUER'} =~ s/,/\//g;
         $tmp->{'ISSUER'} =~ s/\/ /\//g;
         $tmp->{'ISSUER'} =~ s/^\///;
      } elsif ($_ =~ /Last Update.*: (.+)/i) {
         $tmp->{'LAST_UPDATE'} = $1;
      } elsif ($_ =~ /Next Update.*: (.+)/i) {
         $tmp->{'NEXT_UPDATE'} = $1;
      } 
   }   

   # get revoked certs
   $tmp->{'LIST'} = [];
   for($i = 0; $lines[$i] !~ /^[\s\t]*Revoked Certificates:$/i; $i++) {
      $self->{'CACHE'}->{$file} = $tmp;
      return($tmp) if ($lines[$i] =~ /No Revoked Certificates/i);
   }
   $i++;

   while($i < @lines) {
      if($lines[$i] =~ /Serial Number.*: (.+)/i) {
         my $t= {};
         $t->{'SERIAL'} = length($1)%2?"0".uc($1):uc($1);
         $i++;
         if($lines[$i] =~ /Revocation Date: (.*)/i ) {
            $t->{'DATE'} = $1;
            $i++;
            push(@{$tmp->{'LIST'}}, $t);
         } else {
            $main->print_warning("CRL seems to be corrupt: $file\n");
            return;
         }
         
      } else {
         $i++;
      }
   }

   $self->{'CACHE'}->{$file} = $tmp;

   return($tmp);
}

sub parsecert {
   my $self  = shift;
   my $main  = shift;
   my $file  = shift;
   my $force = shift;

   my $tmp   = {};
   my (@lines, @dn, $i, $c, $v, $k, $cmd, $crl, $time);

   my $ca  = $main->{'CA'}->{'actca'};

   $time = time();

   $force && delete($self->{'CACHE'}->{$file});

   # check if certificate is cached
   if($self->{'CACHE'}->{$file}) {
      #  print "DEBUG: use cached certificate\n";
      return($self->{'CACHE'}->{$file});
   }
   #print "DEBUG: parse certificate\n";

   open(IN, $file) || do {
      $main->print_warning("Can't open cert: $!");
      return;
   };

   # convert certificate to PEM, DER and TEXT
   $tmp->{'PEM'} .= $_ while(<IN>);
   $tmp->{'TEXT'} = $self->_convcert(
         $main,
         $tmp->{'PEM'},
         'TEXT'
         );
   $tmp->{'DER'} = $self->_convcert(
         $main,
         $tmp->{'PEM'},
         'DER'
         );

   # get "normal infos"
   @lines = split(/\n/, $tmp->{'TEXT'});
   foreach(@lines) {
      if($_ =~ /Serial Number.*: (.+) /i) {
         $tmp->{'SERIAL'} = length($1)%2?"0".uc($1):uc($1);
      } elsif ($_ =~ /Signature Algorithm.*: (\w+)/i) {
         $tmp->{'SIG_ALGORITHM'} = $1;
      } elsif ($_ =~ /Issuer: (.+)/i) {
         $tmp->{'ISSUER'} = $1;
         $tmp->{'ISSUER'} =~ s/,/\//g;
         $tmp->{'ISSUER'} =~ s/\/ /\//g;
         $tmp->{'ISSUER'} =~ s/^\///;
      } elsif ($_ =~ /Not Before.*: (.+)/i) {
         $tmp->{'NOTBEFORE'} = $1;
      } elsif ($_ =~ /Not After.*: (.+)/i) {
         $tmp->{'NOTAFTER'} = $1;
      } elsif ($_ =~ /Public Key Algorithm.*: (.+)/i) {
         $tmp->{'PK_ALGORITHM'} = $1;
      } elsif ($_ =~ /Modulus \((\d+) .*\)/i) {
         $tmp->{'KEYSIZE'} = $1;
      } elsif ($_ =~ /Subject.*: (.+)/i) {
         $tmp->{'DN'} = $1;
      }
   }   

   $tmp->{'DN'} =~ s/,/\//g;
   @dn = split(/\//, $tmp->{'DN'});
   foreach(@dn) {
      s/^\s+//;
      s/\s+$//;
      ($k, $v) = split(/=/);
      if($k =~ /ou/i) {
         $tmp->{'OU'} or  $tmp->{'OU'} = [];
         push(@{$tmp->{'OU'}}, $v);
      } else {
         $tmp->{uc($k)} = $v;
      }
   }
   

   # get extensions
   $tmp->{'EXT'} = {};
   for($i = 0; $lines[$i] !~ /^[\s\t]*X509v3 extensions:$/i; $i++) {
      return($tmp) if not defined($lines[$i]);
   }
   $i++;

   while($i < @lines) {
      if(($lines[$i] =~ /^[\s\t]*[^:]+:\s*$/) ||
         ($lines[$i] =~ /^[\s\t]*[^:]+:\s+.+$/)) {
         if($lines[$i] =~ /^[\s\t]*Signature Algorithm/i) {
            $i++;
            next;
         }
         $k = $lines[$i];
         $k =~ s/[\s\t:]*$//g;
         $k =~ s/^[\s\t]*//g;
         $tmp->{'EXT'}->{$k} = [];
         $i++;
         while(($lines[$i] !~ /^[\s\t].+:\s*$/) && 
               ($lines[$i] !~ /^[\s\t]*[^:]+:\s+.+$/) &&
               ($lines[$i] !~ /^[\s\t]*Signature Algorithm/i) &&
               ($i < @lines)) {
            $v = $lines[$i];
            $v =~ s/^[\s]+//g;
            $v =~ s/[\s]+$//g;
            $i++;
            next if $v =~ /^$/;
            next if $v =~ /Signature Algorithm:/;
            push(@{$tmp->{'EXT'}->{$k}}, $v);
         }
      } else {
         $i++;
      }
   }

   # get fingerprint 
   $cmd = "$self->{'bin'} x509 -noout -fingerprint -in $file";
   open(CMD, "$cmd|");
   ($k, $v) = split(/=/, <CMD>);
   close(CMD);
   $tmp->{'FINGERPRINT'} = $v if($k =~ /MD5 Fingerprint/i);

   $tmp->{'EXPDATE'} = _get_date( $tmp->{'NOTAFTER'});

   $crl = $self->parsecrl( 
         $main,
         $main->{'CA'}->{$ca}->{'dir'}."/crl/crl.pem"
         );

   defined($crl) || $main->print_error("Can't read Revocation List");

   $tmp->{'STATUS'} = "VALID";

   if($tmp->{'EXPDATE'} < $time) {
      $tmp->{'STATUS'} = "EXPIRED";
      # keep database up to date
      if($crl->{'ISSUER'} eq $tmp->{'ISSUER'}) {
         _set_expired($tmp->{'SERIAL'}, $main);
      }
   }
   
   if(($crl->{'ISSUER'} ne $tmp->{'ISSUER'}) || 
      (not defined($tmp->{'ISSUER'}))) {
      print "DEBUG: Issuers differ:\n"
           ."   crl-issuer: $crl->{'ISSUER'}\n"
           ."   crt-issuer: $tmp->{'ISSUER'}\n";
   
      $tmp->{'STATUS'} = "UNKNOWN";
   }
   
   foreach my $revoked (@{$crl->{'LIST'}}) {
      next if ($tmp->{'SERIAL'} ne $revoked->{'SERIAL'});
      if ($tmp->{'SERIAL'} == $revoked->{'SERIAL'}) {
         $tmp->{'STATUS'} = "REVOKED";
      }
   }

   $self->{'CACHE'}->{$file} = $tmp;

   return($tmp);
}

sub parsereq {
   my $self = shift;
   my $main = shift;
   my $file = shift;

   my $tmp   = {};
   my (@lines, @dn, $i, $c, $v, $k, $cmd);

   # check if request is cached
   if($self->{'CACHE'}->{$file}) {
      return($self->{'CACHE'}->{$file});
   }

   open(IN, $file) || do {
      $main->print_warning("Can't open request: $!");
      return;
   };

   # convert request to PEM, DER and TEXT
   $tmp->{'PEM'} .= $_ while(<IN>);
   $tmp->{'TEXT'} = $self->convreq(
         $main,
         $tmp->{'PEM'},
         'TEXT',
         'PEM'
         );
   $tmp->{'DER'} = $self->convreq(
         $main,
         $tmp->{'PEM'},
         'DER',
         'PEM'
         );

   # get "normal infos"
   @lines = split(/\n/, $tmp->{'TEXT'});
   foreach(@lines) {
      if ($_ =~ /Signature Algorithm.*: (\w+)/i) {
         $tmp->{'SIG_ALGORITHM'} = $1;
      } elsif ($_ =~ /Public Key Algorithm.*: (.+)/i) {
         $tmp->{'PK_ALGORITHM'} = $1;
      } elsif ($_ =~ /Modulus \((\d+) .*\)/i) {
         $tmp->{'KEYSIZE'} = $1;
      } elsif ($_ =~ /Subject.*: (.+)/i) {
         $tmp->{'DN'} = $1;
      }
   }   

   $tmp->{'DN'} =~ s/,/\//g;
   @dn = split(/\//, $tmp->{'DN'});
   foreach(@dn) {
      s/^\s+//;
      s/\s+$//;
      ($k, $v) = split(/=/);
      if($k =~ /ou/i) {
         $tmp->{'OU'} or  $tmp->{'OU'} = [];
         push(@{$tmp->{'OU'}}, $v);
      } else {
         $tmp->{uc($k)} = $v;
      }
   }

   # get extensions
   $tmp->{'EXT'} = {};
   for($i = 0; $lines[$i] !~ /^[\s\t]*Requested extensions:$/i; $i++) {
      return($tmp) if not defined($lines[$i]);
   }
   $i++;

   while($i < @lines) {
      if($lines[$i] =~ /^[\s\t]*[^:]+:\s*$/) {
         $k = $lines[$i];
         $k =~ s/[\s\t:]*$//g;
         $k =~ s/^[\s\t]*//g;
         $tmp->{'EXT'}->{$k} = [];
         $i++;
         while($lines[$i] !~ /^[\s\t].+:\s*$/ && $i < @lines) {
            $v = $lines[$i];
            $v =~ s/^[\s]+//g;
            $v =~ s/[\s]+$//g;
            $i++;
            next if $v =~ /^$/;
            next if $v =~ /Signature Algorithm:/;
            push(@{$tmp->{'EXT'}->{$k}}, $v);
         }
      } else {
         $i++;
      }
   }

   $self->{'CACHE'}->{$file} = $tmp;

   return($tmp);
}

sub _convcrl {
   my $self = shift;
   my $main = shift;
   my $in   = shift;
   my $of   = shift;

   my $tmp  = '';
   my $file = _mktmp($self->{'tmp'}."/cert");

   my $cmd = "$self->{'bin'} crl";
   $cmd .= " -inform PEM";
   $cmd .= " -out $file";
   if($of eq "TEXT") {
      $cmd .= " -text -noout";
   } else {
      $cmd .= " -outform $of";
   }

   open(CMD, "|$cmd");
   print CMD $in;
   close CMD;

   open(IN, $file) || do {
      $main->print_warning("Can't open cert: $!");
      return;
   };
   $tmp .= $_ while(<IN>);
   close(IN);

   unlink($file);

   return($tmp);
}

sub _convcert {
   my $self = shift;
   my $main = shift;
   my $in   = shift;
   my $of   = shift;

   my $tmp  = '';
   my $file = _mktmp($self->{'tmp'}."/cert");

   my $cmd = "$self->{'bin'} x509";
   $cmd .= " -inform PEM";
   $cmd .= " -out $file";
   if($of eq "TEXT") {
      $cmd .= " -text -noout";
   } else {
      $cmd .= " -outform $of";
   }

   open(CMD, "|$cmd");
   print CMD $in;
   close CMD;

   open(IN, $file) || do {
      $main->print_warning("Can't open cert: $!");
      return;
   };
   $tmp .= $_ while(<IN>);
   close(IN);

   unlink($file);

   return($tmp);
}

sub convreq {
   my $self = shift;
   my $main = shift;
   my $in   = shift;
   my $of   = shift;
   my $inform = shift;

   my $tmp  = '';
   my $file = _mktmp($self->{'tmp'}."/req");

   my $cmd = "$self->{'bin'} req";
   $cmd .= " -inform $inform";
   $cmd .= " -out $file";
   if($of eq "TEXT") {
      $cmd .= " -text -noout";
   } else {
      $cmd .= " -outform $of";
   }

   open(CMD, "|$cmd");
   print CMD $in;
   close CMD;

   open(IN, $file) || do {
      $main->print_warning("Can't open req: $!");
      return;
   };
   $tmp .= $_ while(<IN>);
   close(IN);

   unlink($file);

   return($tmp);
}

sub convkey {
   my $self = shift;
   my $main = shift;
   my $input   = shift;
   my $outform = shift;
   my $inform  = shift;
   my $nopass  = shift;
   my $passwd  = shift;

   my $tmp  = '';
   my $file = _mktmp($self->{'tmp'}."/key");

   my $cmd = "$self->{'bin'} rsa";
   $cmd .= " -inform $inform";
   $cmd .= " -outform $outform";
   $cmd .= " -out $file";

   $cmd .= " -passin env:SSLPASS";
   $cmd .= " -passout env:SSLPASSOUT -des3" if(not $nopass);

   $ENV{'SSLPASS'} = $passwd;
   $ENV{'SSLPASSOUT'} = $passwd if(not $nopass);
   
   open(CMD, "|$cmd");
   print CMD $input;
   close CMD;

   delete($ENV{'SSLPASS'});
   delete($ENV{'SSLPASSOUT'});

   open(IN, $file) || return(undef);

   $tmp .= $_ while(<IN>);
   close(IN);

   unlink($file);

   return($tmp);
}

sub genp12 {
   my $self = shift;
   my $opts = { @_ };

   my($main, $cmd);
   
   $main = $opts->{'main'};

   $cmd = "$self->{'bin'} pkcs12 -export";
   $cmd .= " -out $opts->{'outfile'}";
   $cmd .= " -in $opts->{'certfile'}";
   $cmd .= " -inkey $opts->{'keyfile'}";
   $cmd .= " -passout env:P12PASS";
   $cmd .= " -passin env:SSLPASS";
   $cmd .= " -certfile $opts->{'cafile'}" if($opts->{'includeca'});

   $ENV{'P12PASS'} = $opts->{'p12passwd'};
   $ENV{'SSLPASS'} = $opts->{'passwd'};
   system($cmd);
   delete($ENV{'P12PASS'});
   delete($ENV{'SSLPASS'});
}

sub _mktmp { 
   my $base = shift;
   my @rand = ();
   my $ret  = '';

   do { 
      for(my $i = 0; $i < 8; $i++) { 
         push(@rand, int(rand 26)+65);
      }
      my $end = pack("C8", @rand);
      $ret = $base.$end;
   } while (-e $ret);

   return($ret);
}

sub _set_expired {
   my $serial = shift;
   my $main   = shift;
   
   my $ca = $main->{'CA'}->{'actca'};

   my $index = $main->{'CA'}->{$ca}->{'dir'}."/index.txt";

   open(IN, "<$index") || do {
      $main->print_warning("Can't open Index");
      return;
   };

   my @lines = <IN>;

   close IN;

   open(OUT, ">$index") || do {
      $main->print_warning("Can't open Index");
      return;
   };

   foreach my $l (@lines) {
      if($l =~ /\t$serial\t/) {
         $l =~ s/^V/E/;
      }
      print OUT $l;
   }

   close OUT;

   return;
}

sub _get_date {
   my $string = shift;
         
   $string =~ s/  / /g;
            
   my @t1 = split(/ /, $string);
   my @t2 = split(/:/, $t1[2]);

   $t1[0] = _get_index($t1[0]);
                              
   my $ret = Time::Local::timelocal($t2[2],$t2[1],$t2[0],$t1[1],$t1[0],$t1[3]);

   return($ret);
}
   
sub _get_index {
   my $m = shift;
   my @a = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);

   for(my $i = 0; $a[$i]; $i++) {
      return $i if($a[$i] eq $m);
   }
}





   
1
